/**********************************************************************

	Command:  CDVIEW -- Create Clipped Detail View of drawing.

	Programmed by Kang, Hsin-Min.
	Release at Nov 20, 1995 As a working example
	Revised at Jun 25, 1997 to take advantage of BTRIM commands
	Revised at Feb 08, 1998 to avoid a BTRIM bug, to release as
				standard command.
	Revised at Apr 22, 1998 Include Japenese Text in SJIS codes
	Revised at Aug 05, 1998 Modify Japenese Messages.

	Copyright (c) 1995, 1997, 1998 TCAM Development House, Taipei
	All rights Reserved.

	Status: TwinCAD Standard Command, with full support from
		TCAM Development House.

    Comments:	This program can be further improved such that
		un-trimable objects like Texts, Dimensions and
		Block instances are considered.

*********************************************************************/
#define LoopForever	while(1)
#define MidPoint(p1,p2) (p1+p2)/2
#define T_CIRCLE	4	/* Resolved from tclcad.h */
#define FPOLY		1
#define SELMASK_ON	0x2000

#define TGBIG5		1
#define TGGB2312	2
#define TGSJIS		5
#define TGKOREAN	3

int	IsWin;
int	IsNew=identify("BTRIM");// See if we have the support here
				// Litemode may not have BTRIM support
char	szText[5][70];		// Our display text buffer
char	szAsciiText[][60]={	// Default English Display Text
/*  0 */   "\nClipped Details View -- Circle-type/<Box-type>: ",
/*  1 */   "\n** Please box the part of drawing for details view.",
/*  2 */   "\n** Please Circle the part of drawing for details view.",
/*  3 */   "\nDetails View -- <Scale(%@0&f)>/<View Center Location>: ",
/*  4 */   "\nNeed a Corner Point of box, Space-bar or Keyword.",
};

char	szBig5Text[][60]={	// Default Chinese Big-5 Display Text
/*  0 */   "\nӳj -- :ꫬ/<x:Ĥ@I>: ",
/*  1 */   "\n** бNnӳjɳίxήذ_..",
/*  2 */   "\n** бNnӳjɳζذ_...",
/*  3 */   "\nӳj -- <j(%@0&f)>/<Ϥߦm>: ",
/*  4 */   "\nеwI, ťξAƫOrX.",
};

char	szGB2312Text[][60]={	// Default Chinese GB2312 Display Text
/*  0 */   "\nϸŴͼ -- :Բ/<:һ>: ",
/*  1 */   "\n** 뽫ҪϸŴͼþο..",
/*  2 */   "\n** 뽫ҪϸŴͼԲ...",
/*  3 */   "\nϸŴͼ -- <Ŵ(%@0&f)>/<ͼλ>: ",
/*  4 */   "\n, հ׼ʵĸָ.",
};
//
//	The following Japenese messages are translated by _____________.
//
char	szSJISText[][60]={    // Default Japenese SJIS Display Text
/*  0 */   "\nڍו\\ -- C:~`/<`>: ",
/*  1 */   "\n** `̈̏ڍו\\",
/*  2 */   "\n** ~`̈̏ڍו\\",
/*  3 */   "\nڍו\\ -- <(%@0&f)>/<\\̒S>: ",
/*  4 */   "\n`̺Ű̓_Kv, ߰ͷܰ.",
};
//
//	The following Korean messages are translated by _____________.
//
char	szKoreanText[][60]={	// Default Korean Display Text
/*  0 */   "\nClipped Details View -- Circle-type/<Box-type>: ",
/*  1 */   "\n** Please box the part of drawing for details view.",
/*  2 */   "\n** Please Circle the part of drawing for details view.",
/*  3 */   "\nDetails View -- <Scale(%@0&f)>/<View Center Location>: ",
/*  4 */   "\nNeed a Corner Point of box, Space-bar or Keyword.",
};

COMMAND CDView()
{
	int	i;

	IsWin = 1;
	switch(1) {  // Test if we are in TwinCAD ...
	  case !identify("os_type"):
	  case (os_type()<0):	// Must be in DOS version!
		IsWin = 0;
		break;
	}

	DoubleCheck();

	switch(@tgftype) {
	  case TGBIG5:	  setdata(szText,szBig5Text);	  break;
	  case TGGB2312:  setdata(szText,szGB2312Text);   break;
	  case TGSJIS:	  setdata(szText,szSJISText);	  break;
	  case TGKOREAN:  setdata(szText,szKoreanText);   break;
	  default:	  setdata(szText,szAsciiText);	  break;
	}

	set_drag(-1);	// Turn off dragging, if it has been enabled.
	LoopForever {
	    i = getinput(szText[0],0,0,"C");
	    switch(i) {
	      case 0:	    // Space-bar, default to box type
		do_box(0);	break;
	      case 1:	    // Value, bad!
		printf(szText[4]);  continue;
	      case 2:	    // Point, default to box type, first corner
	      case 3:	    // Point, with an associated object
		do_box(1);	break;
	      case 4:	    // Keyword C, choose Circle type
		do_circle();	break;
	    }
	    exit();
	}
}
//
//	We need to double check if the BTRIM command is actually supported.
//	TwinCAD supports Lite Mode since 3.1.044 and BTRIM may not be
//	supported on certain level of lite mode operation. However,
//	the identify("BTRIM") will always return true even the BTRIM is
//	not supported in the Lite Mode, until V3.1.050. So, we need to
//	check further to see if the BTRIM is truely supported.
//
//	Note also that there is a bug in BTRIM before TwinCAD V3.1.50
//	and before its DOS version V3R4C3, which would fail to trim off
//	selected objects after polylines if there is only one boundary
//	object of simple entity.
//
void	DoubleCheck()
{
	if ( !BtrimOkay() ) return;
	if ( IsWin ) {
		if ( version() <= 310 )
			IsNew = -1;	// Need to avoid the bug
	}
	else if ( version() <= 303 )
		IsNew = -1;		// Need to avoid the bug
	return;
}

int	BtrimOkay()
{
	int	level;
	if ( !IsNew ) return IsNew;
	if ( !IsWin ) return IsNew;
	if ( !identify("SystemMode") ) return IsNew;
	level = SystemMode();
	if ( level >= 0 ) return IsNew; // TestDrive or Standard ...
	level = -level; 	// Lite mode
	if ( level >= 6 ) return IsNew;
	IsNew = 0;	// No Support
	return IsNew;
}

ENTITY	eOldFrame;	// Frame on part drawing
ENTITY	eNewFrame;	// Frame for details view
POINT	pBase;		// Base point of part drawing
POINT	pInsert;	// Insert point of details view
double	dScale=1.;	// Scale factor
//
//	do_box -- Process Box Case
//
do_box(int hasp0)
{
	POINT	Pc1, Pc2;
	SLIST	ssOrg, ssNew, ssResult;

	printf(szText[1]);	// Notice to box the part of drawing
	if ( !command(hasp0?"BOX p0":"BOX") )
	    return 0;		// Is the box created ?

	eOldFrame = ent_last(); // Obtain the box, and corner points

	Pc1 = L(ent_pnext(eOldFrame)).pe;
	Pc2 = L(ent_plast(eOldFrame)).ps;
	pBase = MidPoint(Pc1,Pc2);

	AskDetails();	// Deterimine Scale/Position
	EnterSetup();

	p1 = (Pc1-Pc2)*1.05+Pc2;
	p2 = (Pc2-Pc1)*1.05+Pc1;
	command("ZW p1 p2  ");  // Make sure all in window

	bx_sel(Pc1,Pc2,1);	// Select object Crossed by Box
	command("ZP");
	ssOrg=getesel(-1);
	if ( !ent_count(ssOrg) ) {
	    printf("\nNothing to do in the box!");
	    ExitSetup();
	    return 0;
	}

	if ( version() >= 301 ) {	// New version!
	    set_trns(pBase,pInsert-pBase,p(dScale,dScale),0);
	    ssNew = ent_copy(ssOrg);
	}
	else {
	    e0 = ent_last();
	    command("COPY \@ssOrg  \@pBase \@pInsert  ");
	    ssNew = getesel(ent_next(e0));
	    command("SCALE \@ssNew  \@pInsert \@dScale  ");
	}

	Pc1 = L(ent_pnext(eNewFrame)).pe;
	Pc2 = L(ent_plast(eNewFrame)).ps;
	p1 = (Pc1-Pc2)*1.05+Pc2;
	p2 = (Pc2-Pc1)*1.05+Pc1;
	command("ZW p1 p2  ");

	bx_trim(eNewFrame,Pc1,Pc2);
	bx_sel(Pc1,Pc2,0);

	ssResult=getesel(-1);
	command("ERASE a \@ssNew r \@ssResult   ");
	command("ZP");

	ExitSetup();
	return;
}
//
//	do_circle -- Process Circle Case
//
do_circle()
{
	double	rad, dist;
	SLIST	ssOrg, ssNew, ssResult;

	printf(szText[2]);
	if ( !command("CIRCLE") )
	    return 0;
	eOldFrame = ent_last();
	pBase = C(eOldFrame).pc;
	AskDetails();	// Deterimine Scale/Position
	EnterSetup();

	C1 = eOldFrame;
	rad = c1.rad;
	dist = rad*2.2;
	command("ZOOM c c1.pc \@dist  ");
	cx_sel(eOldFrame);
	command("ZP");
	ssOrg=getesel(-1);
	if ( !ent_count(ssOrg) ) {
	    printf("\nNothing to do in the circle!");
	    ExitSetup();
	    return 0;
	}
	if ( version() >= 301 ) {	// New version!
	    set_trns(pBase,pInsert-pBase,p(dScale,dScale),0);
	    ssNew = ent_copy(ssOrg);
	}
	else {
	    e0 = ent_last();
	    command("COPY \@ssOrg  \@pBase \@pInsert  ");
	    ssNew = getesel(ent_next(e0));
	    command("SCALE \@ssNew  \@pInsert \@dScale  ");
	}

	C1 = eNewFrame;
	rad = C(eNewFrame).rad;
	dist = rad*2.5;
	command("ZOOM c c1.pc \@dist  ");

	cx_trim(eNewFrame);
	cx_sel(eNewFrame);
	ssResult=getesel(-1);
	command("ERASE a \@ssNew r \@ssResult   ");
	command("ZP");
	ExitSetup();
	return;
}
//
//	Determine the insert point and scale factor.
//
AskDetails()
{
	SLIST	ssFrame;
	int	i;
	double	CurScale;
	POINT	CurPos;

	set_trns(0);	// Turn off transformation, just in case!
	ssFrame = ent_copy(eOldFrame);
	eNewFrame = ent_last();

	dScale = CurScale = 1;
	pInsert = CurPos = pBase;

	userbrk(0);	// Turn off User Break

	LoopForever {
	    set_drag(0x2060,pInsert,0,0,ssFrame);
	    i = getinput(prints(szText[3],dScale),0,0);
	    set_drag(-1);  set_map(-1);
	    switch(i) {
	      case 0:	return 0;	// Space-Bar, Ok!
	      case 1:		// Value input
		dScale = v0;
		v0 = dScale/CurScale;
		set_trns(pInsert,p(0,0),p(v0,v0),0);
		ent_trns(ssFrame);
		CurScale = dScale;
		set_trns(0);
		redraw();
		continue;
	      case 2:
	      case 3:		// Point input
		pInsert = P0;
		set_trns(pInsert,pInsert-CurPos,p(1.,1.),0);
		ent_trns(ssFrame);
		set_trns(0);
		CurPos = pInsert;
		redraw();
		continue;
	      default:
		command("U");   // Undo last section
		exit();
	   }
	}
}
//
//	bx_sel -- Select objects bounded/Crossed by box entity
//
int	bx_sel(Pc1,Pc2,cross)
POINT	Pc1;		// Corner 1 of box
POINT	Pc2;		// Corner 2 of box
int	cross;		// Cross or Bounded only
{
	if ( cross )
	    command("SELECT A c \@Pc1 \@Pc2    ");
	else
	    command("SELECT A w \@Pc1 \@Pc2    ");
}
//
//	bx_trim -- Trim objects bounded by box entity
//
int	bx_trim(ent,Pc1,Pc2)
ENTITY	ent;		// Box Entity
POINT	Pc1;		// Corner 1 of box
POINT	Pc2;		// Corner 2 of box
{
	if ( IsNew ) {
	    e1 = ent;
	    p1 = L(Pc1,Pc2),-1.;
	    @autoredraw = 1;
	    command("BTRIM A CP B e1 ^M e1^M R P1 ");
	    return 1;
	}
	p1 = L(Pc1,Pc2),-0.001;
	p3 = L(Pc2,Pc1),-0.001;
	p2 = p1.x,p3.y;
	p4 = p3.x,p1.y;
	command("TRIM \@ent  c p1 p2 p3 p4 p1 ^M ");
}
//
//	cx_sel -- Select objects Bounded/Crossed by circle entity
//
int	cx_sel(ent)
ENTITY	ent;		// Circle Entity
{
	int	i;

	if ( !ent_ok(ent) || ent_type(ent)!=T_CIRCLE )
	    return 0;	// No Selection, if bad argument!
	if ( IsNew || IsWin ) {  // Use CrossPolygon/Boundary/eXclusive
	    e1 = ent;
	    command("SELECT A CP B X e1 ^M  ");
	    return 1;
	}
	c1 = ent;	// Read the circle
	for(i=0;i<=19;i++)
	    p[i]=c1,(i*18)A;
	command("SELECT A C p1 p11 C p2 p12 C p3 p13 C p4 p14",
		"f p0 p1 p2 p3 p4 p5 p6 p7 p8 p9 p10 p11 p12",
		"p13 p14 p15 p16 p17 p18 p19 p0  r \@ent   ");
	return 1;
}
//
//	cx_trim -- Trim objects bounded by circle entity
//
#define INCREMENT	2
int	cx_trim(ent)
ENTITY	ent;	// Circle Entity
{
	int	i, j;
	if ( !ent_ok(ent) || ent_type(ent)!=T_CIRCLE )
	    return 0;	// No Trimming, if bad argument!
	c1 = ent;	// Read the circle
	if ( IsNew ) {
	    e1 = ent;
	    c1.rad *= 1.5;
	    p1 = c1,0A;
	    if ( IsNew < 0 ) {
		@autoredraw = 1;    // To avoid an internal display bug
		@selmask = ~FPOLY;
		@snapflag |= SELMASK_ON;
		command("BTRIM A CP B e1 ^M e1^M R P1 ");
		@selmask = FPOLY;
		@snapflag |= SELMASK_ON;
	    }
	    command("BTRIM A CP B e1 ^M e1^M R P1 ");
	    return 1;
	}

	printf("\nI am working..");
	c1.rad /= cos(9); // Find circumscribing polygon size
	c1.rad *= 1.0001; // Avoid triming exactly at tangent point
	for(j=0;j<18;j+=INCREMENT) {
	    printf(".");
	    for(i=0;i<=19;i++)
		p[i]=c1,(i*18+j)A;
	    command("TRIM \@ent  c p0 p1 p2 p3 p4 p5 p6 p7 p8 p9 p10 p11 p12",
		    "p13 p14 p15 p16 p17 p18 p19 p0   ");
	}
	printf(" really hard!");
	return 1;
}
//
//	EnterSetup -- Entering Program Setup
//
int	cmdecho=@cmdecho;
int	blipmode=@blipmode;
int	autoredraw=@autoredraw;
int	use_emode=@use_emode;
int	selmask=@selmask;
int	highlight;
int	snapflag=@snapflag;
EnterSetup()
{
	@cmdecho=0;
	@blipmode=0;
	@autoredraw = 0;
	@use_emode=0;
	if ( version() >= 0x301 ) {
		highlight=@highlight;
		@highlight = 0;
	}
	@snapflag &= ~SELMASK_ON;	// Turn off object selection mask
	set_undo(0);	// Specify to regen when undo
}
//
//	ExitSetup -- Exit Program Setup, Restore System Original
//
ExitSetup()
{
	@cmdecho = cmdecho;
	@blipmode = blipmode;
	@autoredraw=autoredraw;
	@use_emode = use_emode;
	if ( version() >= 0x301 )
		@highlight = highlight;
	@snapflag = snapflag;
	@selmask = selmask;
}
